﻿/*
  KPUInput - KeePass Plugin
  Copyright (C) 2020-2022 Dominik Reichl <dominik.reichl@t-online.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Threading;

using KeePass.Plugins;
using KeePass.Util;
using KeePass.Util.SendInputExt;

using KeePassLib.Native;

namespace KPUInput
{
	public sealed class KPUInputExt : Plugin
	{
		// The initialization of a SiEngineUInput object takes
		// some time, therefore we prepare one in advance
		private static SiEngineUInput g_siNext = null;
		private static readonly object g_oNextSync = new object();

		private IPluginHost m_host = null;

		public override bool Initialize(IPluginHost host)
		{
			if(!NativeLib.IsUnix()) return false;

			Terminate();
			if(host == null) { Debug.Assert(false); return false; }
			m_host = host;

			SendInputEx.CreateEngine += KPUInputExt.CreateEngine;

			PrepareNextEngine();
			return true;
		}

		public override void Terminate()
		{
			if(m_host == null) return;

			SendInputEx.CreateEngine -= KPUInputExt.CreateEngine;

			m_host = null;
		}

		private static void CreateEngine(object sender, SiEventArgs e)
		{
			if((e == null) || (e.Engine != null)) return;

			SiEngineUInput si;
			lock(g_oNextSync) { si = g_siNext; g_siNext = null; }

			// The following should happen only when auto-type
			// invokes itself (recursively)
			if(si == null) { Debug.Assert(false); si = new SiEngineUInput(); }

			si.Released += KPUInputExt.OnEngineReleased;

			e.Engine = si;
		}

		private static void PrepareNextEngine()
		{
			try
			{
				ThreadPool.QueueUserWorkItem(new WaitCallback(
					KPUInputExt.PrepareNextEngineTh));
			}
			catch(Exception) { Debug.Assert(false); }
		}

		private static void PrepareNextEngineTh(object state)
		{
			try
			{
				SiEngineUInput si = new SiEngineUInput();
				lock(g_oNextSync) { g_siNext = si; }
			}
			catch(Exception) { Debug.Assert(false); }
		}

		private static void OnEngineReleased(object sender, EventArgs e)
		{
			try
			{
				PrepareNextEngine();

				SiEngineUInput si = (sender as SiEngineUInput);
				if(si != null) si.Released -= KPUInputExt.OnEngineReleased;
				else { Debug.Assert(false); }
			}
			catch(Exception) { Debug.Assert(false); }
		}
	}
}
