﻿/*
  KPUInput - KeePass Plugin
  Copyright (C) 2020-2022 Dominik Reichl <dominik.reichl@t-online.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Windows.Forms;

using KeePass.Util.SendInputExt;

namespace KPUInput
{
	public class SiEngineUInput : ISiEngine, IDisposable
	{
		private const string KPUInputDll = "KPUInputN.so";

		[DllImport(KPUInputDll)]
		private static extern IntPtr KpuiOpenDevice(IntPtr pKeys, uint cKeys);

		[DllImport(KPUInputDll)]
		private static extern void KpuiCloseDevice(IntPtr hDevice);

		[DllImport(KPUInputDll)]
		private static extern void KpuiStartSequence(IntPtr hDevice);

		[DllImport(KPUInputDll)]
		private static extern void KpuiEndSequence(IntPtr hDevice);

		[DllImport(KPUInputDll)]
		private static extern void KpuiSendKey(IntPtr hDevice, ushort uKey, uint uFlags);

		[DllImport(KPUInputDll)]
		private static extern void KpuiSendChar(IntPtr hDevice, uint ch);

		// [DllImport(KPUInputDll, CharSet = CharSet.Ansi)]
		// private static extern void KpuiSetProperty(IntPtr hDevice, string lpKey, string lpValue);

		private const uint KPUI_S_DOWN = 1;
		private const uint KPUI_S_UP = 2;

		private static Dictionary<int, ushort> g_dWToL = null;
		/// <summary>
		/// Translate <c>System.Windows.Forms.Keys</c> to Linux key codes.
		/// https://github.com/torvalds/linux/blob/master/include/uapi/linux/input-event-codes.h
		/// </summary>
		private static Dictionary<int, ushort> MapWToL
		{
			get
			{
				Dictionary<int, ushort> d = g_dWToL;
				if(d == null)
				{
					d = new Dictionary<int, ushort>();

					// d[(int)Keys.Cancel] = 223;
					d[(int)Keys.Back] = 14;
					d[(int)Keys.Tab] = 15;
					d[(int)Keys.LineFeed] = 101;
					d[(int)Keys.Clear] = 0x163; // Cf. Keys.OemClear
					d[(int)Keys.Enter] = 28;
					d[(int)Keys.ShiftKey] = 42;
					d[(int)Keys.ControlKey] = 29;
					d[(int)Keys.Menu] = 56;
					d[(int)Keys.Pause] = 119;
					d[(int)Keys.CapsLock] = 58;
					d[(int)Keys.HangulMode] = 122;
					d[(int)Keys.HanjaMode] = 123;
					d[(int)Keys.Escape] = 1;
					d[(int)Keys.Space] = 57;
					d[(int)Keys.PageUp] = 104;
					d[(int)Keys.PageDown] = 109;
					d[(int)Keys.End] = 107;
					d[(int)Keys.Home] = 102;
					d[(int)Keys.Left] = 105;
					d[(int)Keys.Up] = 103;
					d[(int)Keys.Right] = 106;
					d[(int)Keys.Down] = 108;
					d[(int)Keys.Select] = 0x161;
					d[(int)Keys.Print] = 210;
					d[(int)Keys.PrintScreen] = 99; // KEY_SYSRQ
					d[(int)Keys.Insert] = 110;
					d[(int)Keys.Delete] = 111;
					d[(int)Keys.Help] = 138;

					d[(int)Keys.D0] = 11;
					for(int i = 0; i < 9; ++i)
						d[(int)Keys.D1 + i] = (ushort)(2 + i);

					d[(int)Keys.A] = 30;
					d[(int)Keys.B] = 48;
					d[(int)Keys.C] = 46;
					d[(int)Keys.D] = 32;
					d[(int)Keys.E] = 18;
					d[(int)Keys.F] = 33;
					d[(int)Keys.G] = 34;
					d[(int)Keys.H] = 35;
					d[(int)Keys.I] = 23;
					d[(int)Keys.J] = 36;
					d[(int)Keys.K] = 37;
					d[(int)Keys.L] = 38;
					d[(int)Keys.M] = 50;
					d[(int)Keys.N] = 49;
					d[(int)Keys.O] = 24;
					d[(int)Keys.P] = 25;
					d[(int)Keys.Q] = 16;
					d[(int)Keys.R] = 19;
					d[(int)Keys.S] = 31;
					d[(int)Keys.T] = 20;
					d[(int)Keys.U] = 22;
					d[(int)Keys.V] = 47;
					d[(int)Keys.W] = 17;
					d[(int)Keys.X] = 45;
					d[(int)Keys.Y] = 21;
					d[(int)Keys.Z] = 44;
					d[(int)Keys.LWin] = 125; // KEY_LEFTMETA
					d[(int)Keys.RWin] = 126; // KEY_RIGHTMETA
					d[(int)Keys.Sleep] = 142;
					d[(int)Keys.NumPad0] = 82;
					d[(int)Keys.NumPad1] = 79;
					d[(int)Keys.NumPad2] = 80;
					d[(int)Keys.NumPad3] = 81;
					d[(int)Keys.NumPad4] = 75;
					d[(int)Keys.NumPad5] = 76;
					d[(int)Keys.NumPad6] = 77;
					d[(int)Keys.NumPad7] = 71;
					d[(int)Keys.NumPad8] = 72;
					d[(int)Keys.NumPad9] = 73;
					d[(int)Keys.Multiply] = 55;
					d[(int)Keys.Add] = 78;
					d[(int)Keys.Separator] = 121; // KEY_KPCOMMA
					d[(int)Keys.Subtract] = 74;
					d[(int)Keys.Decimal] = 83; // KEY_KPDOT
					d[(int)Keys.Divide] = 98;

					for(int i = 0; i < 10; ++i)
						d[(int)Keys.F1 + i] = (ushort)(59 + i);
					d[(int)Keys.F11] = 87;
					d[(int)Keys.F12] = 88;
					for(int i = 0; i < 12; ++i)
						d[(int)Keys.F13 + i] = (ushort)(183 + i);

					d[(int)Keys.NumLock] = 69;
					d[(int)Keys.Scroll] = 70;
					d[(int)Keys.LShiftKey] = 42;
					d[(int)Keys.RShiftKey] = 54;
					d[(int)Keys.LControlKey] = 29;
					d[(int)Keys.RControlKey] = 97;
					d[(int)Keys.LMenu] = 56;
					d[(int)Keys.RMenu] = 100;
					d[(int)Keys.BrowserBack] = 158;
					d[(int)Keys.BrowserForward] = 159;
					d[(int)Keys.BrowserRefresh] = 173;
					d[(int)Keys.BrowserStop] = 128;
					d[(int)Keys.BrowserSearch] = 136;
					d[(int)Keys.BrowserFavorites] = 0x16C;
					d[(int)Keys.BrowserHome] = 172;
					d[(int)Keys.VolumeMute] = 113;
					d[(int)Keys.VolumeDown] = 114;
					d[(int)Keys.VolumeUp] = 115;
					d[(int)Keys.MediaNextTrack] = 163;
					d[(int)Keys.MediaPreviousTrack] = 165;
					d[(int)Keys.MediaStop] = 166;
					d[(int)Keys.MediaPlayPause] = 164;
					d[(int)Keys.LaunchMail] = 155;
					d[(int)Keys.OemSemicolon] = 39;
					d[(int)Keys.Oemplus] = 13; // KEY_EQUAL
					d[(int)Keys.Oemcomma] = 51;
					d[(int)Keys.OemMinus] = 12;
					d[(int)Keys.OemPeriod] = 52;
					d[(int)Keys.OemQuestion] = 53; // KEY_SLASH
					d[(int)Keys.OemOpenBrackets] = 26; // KEY_LEFTBRACE
					d[(int)Keys.OemPipe] = 43; // KEY_BACKSLASH
					d[(int)Keys.OemCloseBrackets] = 27; // KEY_RIGHTBRACE
					d[(int)Keys.OemQuotes] = 40; // KEY_APOSTROPHE
					d[(int)Keys.Oem102] = 86; // '<' on German keyboard
					d[(int)Keys.Play] = 207;
					d[(int)Keys.Zoom] = 0x174;
					d[(int)Keys.OemClear] = 0x163; // Cf. Keys.Clear

					g_dWToL = d;
				}

				return d;
			}
		}

		private IntPtr m_h = IntPtr.Zero;

		public event EventHandler Released;

		public SiEngineUInput()
		{
			Dictionary<int, ushort> dMap = SiEngineUInput.MapWToL;
			short[] vKeys = new short[dMap.Count];
			int i = 0;
			foreach(ushort uKey in dMap.Values)
				vKeys[i++] = (short)uKey;
			Debug.Assert(i == vKeys.Length);

			IntPtr pKeys = IntPtr.Zero;
			try
			{
				pKeys = Marshal.AllocCoTaskMem(vKeys.Length * 2);
				if(pKeys == IntPtr.Zero) { Debug.Assert(false); return; }
				Marshal.Copy(vKeys, 0, pKeys, vKeys.Length);

				m_h = KpuiOpenDevice(pKeys, (uint)vKeys.Length);
				Debug.Assert(m_h != IntPtr.Zero);
			}
			finally { if(pKeys != IntPtr.Zero) Marshal.FreeCoTaskMem(pKeys); }
		}

		~SiEngineUInput()
		{
			Dispose(false);
		}

		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		protected virtual void Dispose(bool disposing)
		{
			if(m_h == IntPtr.Zero) return;

			try { KpuiCloseDevice(m_h); }
			catch(Exception) { Debug.Assert(false); }
			finally { m_h = IntPtr.Zero; }
		}

		public void Init()
		{
			if(m_h == IntPtr.Zero) { Debug.Assert(false); return; }

			Keys kMod = Control.ModifierKeys;
			bool bShift = ((kMod & Keys.Shift) != Keys.None);
			bool bCtrl = ((kMod & Keys.Control) != Keys.None);
			bool bAlt = ((kMod & Keys.Alt) != Keys.None);

			// Release modifier keys
			if(bShift || bCtrl || bAlt)
			{
				KpuiStartSequence(m_h);

				if(bShift)
				{
					SendKey((int)Keys.LShiftKey, null, false);
					SendKey((int)Keys.RShiftKey, null, false);
				}
				if(bCtrl)
				{
					SendKey((int)Keys.LControlKey, null, false);
					SendKey((int)Keys.RControlKey, null, false);
				}
				if(bAlt)
				{
					SendKey((int)Keys.LMenu, null, false);
					SendKey((int)Keys.RMenu, null, false);
				}

				KpuiEndSequence(m_h);
			}

			KpuiStartSequence(m_h); // Update state with new modifiers
		}

		public void Release()
		{
			if(m_h == IntPtr.Zero) { Debug.Assert(false); return; }

			KpuiEndSequence(m_h);
			Dispose();

			if(this.Released != null) this.Released(this, EventArgs.Empty);
		}

		public void UpdateExpectedFocus()
		{
		}

		public void SendKey(int iVKey, bool? obExtKey, bool? obDown)
		{
			if(m_h == IntPtr.Zero) { Debug.Assert(false); return; }

			Dictionary<int, ushort> d = SiEngineUInput.MapWToL;
			ushort uKey;
			if(!d.TryGetValue(iVKey, out uKey)) return;

			if(!obDown.HasValue || obDown.Value)
				KpuiSendKey(m_h, uKey, KPUI_S_DOWN);
			if(!obDown.HasValue || !obDown.Value)
				KpuiSendKey(m_h, uKey, KPUI_S_UP);

			Application.DoEvents();
		}

		public void SetKeyModifier(Keys kMod, bool bDown)
		{
			if((kMod & Keys.Shift) != Keys.None)
				SendKey((int)Keys.LShiftKey, null, bDown);
			if((kMod & Keys.Control) != Keys.None)
				SendKey((int)Keys.LControlKey, null, bDown);
			if((kMod & Keys.Alt) != Keys.None)
				SendKey((int)Keys.LMenu, null, bDown);
		}

		public void SendChar(char ch, bool? obDown)
		{
			if(m_h == IntPtr.Zero) { Debug.Assert(false); return; }

			if(char.IsSurrogate(ch)) { Debug.Assert(false); return; }

			if(obDown.GetValueOrDefault(true))
			{
				Debug.Assert(Marshal.SizeOf(typeof(char)) == 2);
				KpuiSendChar(m_h, (uint)(ushort)ch);

				Application.DoEvents();
			}
		}

		public void Delay(uint uMs)
		{
			Thread.Sleep((int)uMs);
		}
	}
}
